/* procmime.c */

#include <glib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "main.h"
#include "procmime.h"
#include "procheader.h"
#include "base64.h"
#include "unmime.h"
#include "codeconv.h"
#include "utils.h"
#include "prefs_common.h"

static GHashTable *procmime_get_mime_type_table	(void);
static GList *procmime_get_mime_type_list	(void);

MimeInfo *procmime_mimeinfo_new(void)
{
	MimeInfo *mimeinfo;

	mimeinfo = g_new0(MimeInfo, 1);
	mimeinfo->mime_type     = MIME_UNKNOWN;
	mimeinfo->encoding_type = ENC_UNKNOWN;

	return mimeinfo;
}

void procmime_mimeinfo_free(MimeInfo *mimeinfo)
{
	if (!mimeinfo) return;

	g_free(mimeinfo->encoding);
	g_free(mimeinfo->content_type);
	g_free(mimeinfo->charset);
	g_free(mimeinfo->name);
	g_free(mimeinfo->boundary);
	g_free(mimeinfo->content_disposition);
	g_free(mimeinfo->filename);

	procmime_mimeinfo_free(mimeinfo->sub);

	g_free(mimeinfo);
}

void procmime_mimeinfo_free_all(MimeInfo *mimeinfo)
{
	while (mimeinfo != NULL) {
		MimeInfo *next;

		g_free(mimeinfo->encoding);
		g_free(mimeinfo->content_type);
		g_free(mimeinfo->charset);
		g_free(mimeinfo->name);
		g_free(mimeinfo->boundary);
		g_free(mimeinfo->content_disposition);
		g_free(mimeinfo->filename);

		procmime_mimeinfo_free_all(mimeinfo->sub);
		procmime_mimeinfo_free_all(mimeinfo->children);

		next = mimeinfo->next;
		g_free(mimeinfo);
		mimeinfo = next;
	}
}

MimeInfo *procmime_mimeinfo_insert(MimeInfo *parent, MimeInfo *mimeinfo)
{
	MimeInfo *child = parent->children;

	if (!child)
		parent->children = mimeinfo;
	else {
		while (child->next != NULL)
			child = child->next;

		child->next = mimeinfo;
	}

	mimeinfo->parent = parent;
	mimeinfo->level = parent->level + 1;

	return mimeinfo;
}

MimeInfo *procmime_scan_message(const gchar *file)
{
	FILE *fp;
	MimeInfo *mimeinfo;

	g_return_val_if_fail(file != NULL, NULL);

	if ((fp = fopen(file, "r")) == NULL) {
		FILE_OP_ERROR(file, "fopen");
		return NULL;
	}

	mimeinfo = procmime_scan_mime_header(fp);

	if (mimeinfo->mime_type == MIME_MULTIPART)
		procmime_scan_multipart_message(mimeinfo, fp);

	fclose(fp);

	return mimeinfo;
}

void procmime_scan_multipart_message(MimeInfo *mimeinfo, FILE *fp)
{
	gchar *p;
	gchar *boundary;
	gint boundary_len;
	gchar buf[BUFFSIZE];
	glong fpos, prev_fpos;
	gint npart;

	g_return_if_fail(mimeinfo != NULL);
	g_return_if_fail(mimeinfo->mime_type == MIME_MULTIPART);
	g_return_if_fail(mimeinfo->boundary != NULL);
	g_return_if_fail(mimeinfo->sub == NULL);
	g_return_if_fail(fp != NULL);

	boundary = mimeinfo->boundary;
	boundary_len = strlen(boundary);

	/* look for first boundary */
	while ((p = fgets(buf, sizeof(buf), fp)) != NULL)
		if (IS_BOUNDARY(buf, boundary, boundary_len)) break;
	if (!p) return;

	if ((fpos = ftell(fp)) < 0) {
		perror("ftell");
		return;
	}

	for (npart = 0;; npart++) {
		MimeInfo *partinfo;
		gboolean eom = FALSE;

		prev_fpos = fpos;

		partinfo = procmime_scan_mime_header(fp);
		if (!partinfo) break;
		procmime_mimeinfo_insert(mimeinfo, partinfo);
		partinfo->fpos = fpos;

		if (partinfo->mime_type == MIME_MULTIPART) {
			if (partinfo->level < 8)
				procmime_scan_multipart_message(partinfo, fp);
		} else if (partinfo->mime_type == MIME_MESSAGE_RFC822) {
			MimeInfo *sub;

			fpos = ftell(fp);
			partinfo->sub = sub = procmime_scan_mime_header(fp);
			if (!sub) break;
			sub->fpos = fpos;
			sub->level = partinfo->level + 1;

			if (sub->mime_type == MIME_MULTIPART) {
				if (sub->level < 8)
					procmime_scan_multipart_message
						(sub, fp);
			}
		}

		/* look for next boundary */
		buf[0] = '\0';
		while ((p = fgets(buf, sizeof(buf), fp)) != NULL) {
			if (IS_BOUNDARY(buf, boundary, boundary_len)) {
				if (buf[2 + boundary_len]     == '-' &&
				    buf[2 + boundary_len + 1] == '-')
					eom = TRUE;
				break;
			}
		}
		if (p == NULL)
			eom = TRUE;	/* broken MIME message */
		fpos = ftell(fp);

		partinfo->size = fpos - prev_fpos - strlen(buf);

		if (eom) break;
	}
}

void procmime_scan_encoding(MimeInfo *mimeinfo, const gchar *encoding)
{
	gchar *buf;

	Xalloca(buf, strlen(encoding) + 1, return);
	strcpy(buf, encoding);

	g_free(mimeinfo->encoding);

	mimeinfo->encoding = g_strdup(g_strstrip(buf));
	if (!strcasecmp(buf, "7bit"))
		mimeinfo->encoding_type = ENC_7BIT;
	else if (!strcasecmp(buf, "8bit"))
		mimeinfo->encoding_type = ENC_8BIT;
	else if (!strcasecmp(buf, "quoted-printable"))
		mimeinfo->encoding_type = ENC_QUOTED_PRINTABLE;
	else if (!strcasecmp(buf, "base64"))
		mimeinfo->encoding_type = ENC_BASE64;
	else
		mimeinfo->encoding_type = ENC_UNKNOWN;

}

void procmime_scan_content_type(MimeInfo *mimeinfo, const gchar *content_type)
{
	gchar *delim, *p, *cnttype;
	gchar *buf;

	Xalloca(buf, strlen(content_type) + 1, return);
	strcpy(buf, content_type);

	g_free(mimeinfo->content_type);
	g_free(mimeinfo->charset);
	g_free(mimeinfo->name);
	mimeinfo->content_type = NULL;
	mimeinfo->charset      = NULL;
	mimeinfo->name         = NULL;

	if ((delim = strchr(buf, ';'))) *delim = '\0';
	mimeinfo->content_type = cnttype = g_strdup(g_strstrip(buf));

	if (!strncasecmp(cnttype, "text/html", 9))
		mimeinfo->mime_type = MIME_TEXT_HTML;
	else if (!strncasecmp(cnttype, "text/", 5))
		mimeinfo->mime_type = MIME_TEXT;
	else if (!strncasecmp(cnttype, "message/rfc822", 14))
		mimeinfo->mime_type = MIME_MESSAGE_RFC822;
	else if (!strncasecmp(cnttype, "message/", 8))
		mimeinfo->mime_type = MIME_TEXT;
	else if (!strncasecmp(cnttype, "application/", 12))
		mimeinfo->mime_type = MIME_APPLICATION;
	else if (!strncasecmp(cnttype, "multipart/", 10))
		mimeinfo->mime_type = MIME_MULTIPART;
	else if (!strncasecmp(cnttype, "image/", 6))
		mimeinfo->mime_type = MIME_IMAGE;
	else if (!strncasecmp(cnttype, "audio/", 6))
		mimeinfo->mime_type = MIME_AUDIO;
	else if (!strcasecmp(cnttype, "text"))
		mimeinfo->mime_type = MIME_TEXT;
	else
		mimeinfo->mime_type = MIME_UNKNOWN;

	if (!delim) return;
	p = delim + 1;

	for (;;) {
		gchar *eq;
		gchar *attr, *value;

		if ((delim = strchr(p, ';'))) *delim = '\0';

		if (!(eq = strchr(p, '='))) break;

		*eq = '\0';
		attr = p;
		g_strstrip(attr);
		value = eq + 1;
		g_strstrip(value);

		if (*value == '"')
			extract_quote(value, '"');
		else {
			eliminate_parenthesis(value, '(', ')');
			g_strstrip(value);
		}

		if (*value) {
			if (!strcasecmp(attr, "charset"))
				mimeinfo->charset = g_strdup(value);
			else if (!strcasecmp(attr, "name")) {
				gchar *tmp;

				UnMimeHeader(value);
				Xalloca(tmp, strlen(value) + 1, return);
				conv_anytoeuc(tmp, strlen(value) + 1, value);
				conv_unreadable_eucjp(tmp);
				mimeinfo->name = g_strdup(tmp);
			} else if (!strcasecmp(attr, "boundary"))
				mimeinfo->boundary = g_strdup(value);
		}

		if (!delim) break;
		p = delim + 1;
	}

	if (mimeinfo->mime_type == MIME_MULTIPART && !mimeinfo->boundary)
		mimeinfo->mime_type = MIME_TEXT;
}

void procmime_scan_content_disposition(MimeInfo *mimeinfo,
				       const gchar *content_disposition)
{
	gchar *delim, *p, *dispos;
	gchar *buf;

	Xalloca(buf, strlen(content_disposition) + 1, return);
	strcpy(buf, content_disposition);

	if ((delim = strchr(buf, ';'))) *delim = '\0';
	mimeinfo->content_disposition = dispos = g_strdup(g_strstrip(buf));

	if (!delim) return;
	p = delim + 1;

	for (;;) {
		gchar *eq;
		gchar *attr, *value;

		if ((delim = strchr(p, ';'))) *delim = '\0';

		if (!(eq = strchr(p, '='))) break;

		*eq = '\0';
		attr = p;
		g_strstrip(attr);
		value = eq + 1;
		g_strstrip(value);

		if (*value == '"')
			extract_quote(value, '"');
		else {
			eliminate_parenthesis(value, '(', ')');
			g_strstrip(value);
		}

		if (*value) {
			if (!strcasecmp(attr, "filename")) {
				gchar *tmp;

				UnMimeHeader(value);
				Xalloca(tmp, strlen(value) + 1, return);
				conv_anytoeuc(tmp, strlen(value) + 1, value);
				conv_unreadable_eucjp(tmp);
				g_free(mimeinfo->filename);
				mimeinfo->filename = g_strdup(tmp);
				break;
			}
		}

		if (!delim) break;
		p = delim + 1;
	}
}

enum
{
	H_CONTENT_TRANSFER_ENCODING = 0,
	H_CONTENT_TYPE		    = 1,
	H_CONTENT_DISPOSITION	    = 2
};

MimeInfo *procmime_scan_mime_header(FILE *fp)
{
	static HeaderEntry hentry[] = {{"Content-Transfer-Encoding:",
							  NULL, FALSE},
				       {"Content-Type:", NULL, TRUE},
				       {"Content-Disposition:",
							  NULL, TRUE},
				       {NULL,		  NULL, FALSE}};
	gchar buf[BUFFSIZE];
	gint hnum;
	HeaderEntry *hp;
	MimeInfo *mimeinfo;

	g_return_val_if_fail(fp != NULL, NULL);

	mimeinfo = procmime_mimeinfo_new();
	mimeinfo->mime_type = MIME_TEXT;
	mimeinfo->encoding_type = ENC_7BIT;

	while ((hnum = procheader_get_one_field(buf, sizeof(buf), fp, hentry))
	       != -1) {
		hp = hentry + hnum;

		if (H_CONTENT_TRANSFER_ENCODING == hnum) {
			procmime_scan_encoding
				(mimeinfo, buf + strlen(hp->name));
		} else if (H_CONTENT_TYPE == hnum) {
			procmime_scan_content_type
				(mimeinfo, buf + strlen(hp->name));
		} else if (H_CONTENT_DISPOSITION == hnum) {
			procmime_scan_content_disposition
				(mimeinfo, buf + strlen(hp->name));
		}
	}

	if (!mimeinfo->content_type)
		mimeinfo->content_type = g_strdup("text/plain");

	return mimeinfo;
}

FILE *procmime_decode_content(FILE *outfp, FILE *infp, MimeInfo *mimeinfo)
{
	gchar buf[BUFFSIZE];
	gchar *boundary = NULL;
	gint boundary_len = 0;
	gboolean tmp_file = FALSE;

	g_return_val_if_fail(infp != NULL, NULL);
	g_return_val_if_fail(mimeinfo != NULL, NULL);

	if (!outfp) {
		outfp = tmpfile();
		if (!outfp) {
			perror("tmpfile");
			return NULL;
		}
		tmp_file = TRUE;
	}

	if (mimeinfo->parent && mimeinfo->parent->boundary) {
		boundary = mimeinfo->parent->boundary;
		boundary_len = strlen(boundary);
	}

	if (mimeinfo->encoding_type == ENC_QUOTED_PRINTABLE) {
		gboolean softline = FALSE;

		while (fgets(buf, sizeof(buf), infp) != NULL &&
		       (!boundary ||
			!IS_BOUNDARY(buf, boundary, boundary_len))) {
			guchar *p = buf;

			softline = DoOneQPLine(&p, FALSE, softline);
			fputs(buf, outfp);
		}
	} else if (mimeinfo->encoding_type == ENC_BASE64) {
		gchar outbuf[BUFFSIZE];
		gint len;

		while (fgets(buf, sizeof(buf), infp) != NULL &&
		       (!boundary ||
			!IS_BOUNDARY(buf, boundary, boundary_len))) {
			len = from64tobits(outbuf, buf);
			if (len < 0) {
				g_warning("Bad BASE64 content\n");
				break;
			}
			fwrite(outbuf, sizeof(gchar), len, outfp);
		}
	} else {
		while (fgets(buf, sizeof(buf), infp) != NULL &&
		       (!boundary ||
			!IS_BOUNDARY(buf, boundary, boundary_len))) {
			fputs(buf, outfp);
		}
	}

	if (tmp_file) rewind(outfp);
	return outfp;
}

gint procmime_get_part(const gchar *outfile, const gchar *infile,
		       MimeInfo *mimeinfo)
{
	FILE *infp, *outfp;
	gchar buf[BUFFSIZE];

	g_return_val_if_fail(outfile != NULL, -1);
	g_return_val_if_fail(infile != NULL, -1);
	g_return_val_if_fail(mimeinfo != NULL, -1);

	if ((infp = fopen(infile, "r")) == NULL) {
		FILE_OP_ERROR(infile, "fopen");
		return -1;
	}
	if (fseek(infp, mimeinfo->fpos, SEEK_SET) < 0) {
		FILE_OP_ERROR(infile, "fseek");
		fclose(infp);
		return -1;
	}
	if ((outfp = fopen(outfile, "w")) == NULL) {
		FILE_OP_ERROR(outfile, "fopen");
		fclose(infp);
		return -1;
	}

	while (fgets(buf, sizeof(buf), infp) != NULL)
		if (buf[0] == '\r' || buf[0] == '\n') break;

	procmime_decode_content(outfp, infp, mimeinfo);

	fclose(infp);
	if (fclose(outfp) == EOF) {
		FILE_OP_ERROR(outfile, "fclose");
		unlink(outfile);
		return -1;
	}

	return 0;
}

FILE *procmime_get_text_part(const gchar *file)
{
	FILE *infp, *tmpfp, *outfp;
	MimeInfo *mimeinfo, *partinfo = NULL;
	CodeConvFunc code_conv = conv_jistodisp;
	gchar buf[BUFFSIZE], buf2[BUFFSIZE];

	g_return_val_if_fail(file != NULL, NULL);

	mimeinfo = procmime_scan_message(file);
	if (!mimeinfo) return NULL;

	if ((infp = fopen(file, "r")) == NULL) {
		FILE_OP_ERROR(file, "fopen");
		procmime_mimeinfo_free_all(mimeinfo);
		return NULL;
	}

	if (mimeinfo->mime_type == MIME_MULTIPART) {
		partinfo = mimeinfo->children;
		if (partinfo && partinfo->mime_type == MIME_TEXT) {
			if (fseek(infp, partinfo->fpos, SEEK_SET) < 0) {
				FILE_OP_ERROR(file, "fseek");
				partinfo = NULL;
			}
		} else
			partinfo = NULL;
	} else if (mimeinfo->mime_type == MIME_TEXT)
		partinfo = mimeinfo;

	if (!partinfo) {
		procmime_mimeinfo_free_all(mimeinfo);
		return NULL;
	}

	while (fgets(buf, sizeof(buf), infp) != NULL)
		if (buf[0] == '\r' || buf[0] == '\n') break;

	tmpfp = procmime_decode_content(NULL, infp, partinfo);
	if (!tmpfp) {
		procmime_mimeinfo_free_all(mimeinfo);
		return NULL;
	}

	if ((outfp = tmpfile()) == NULL) {
		perror("tmpfile");
		fclose(tmpfp);
		procmime_mimeinfo_free_all(mimeinfo);
		return NULL;
	}

	if (prefs_common.force_charset)
		code_conv =
			conv_get_code_conv_func(prefs_common.force_charset);
	else if (partinfo->charset)
		code_conv = conv_get_code_conv_func(partinfo->charset);

	while (fgets(buf, sizeof(buf), tmpfp) != NULL) {
		code_conv(buf2, sizeof(buf2), buf);
		fputs(buf2, outfp);
	}

	fclose(tmpfp);
	procmime_mimeinfo_free_all(mimeinfo);
	rewind(outfp);

	return outfp;
}

static GList *mime_type_list = NULL;

gchar *procmime_get_mime_type(const gchar *filename)
{
	static GHashTable *mime_type_table = NULL;
	MimeType *mime_type;
	const gchar *ext, *p;

	if (!mime_type_table) {
		mime_type_table = procmime_get_mime_type_table();
		if (!mime_type_table) return NULL;
	}

	filename = g_basename(filename);
	p = strrchr(filename, '.');
	if (p)
		ext = p + 1;
	else
		return NULL;

	mime_type = g_hash_table_lookup(mime_type_table, ext);
	if (mime_type) {
		gchar *str;

		str = g_strconcat(mime_type->type, "/", mime_type->sub_type,
				  NULL);
		return str;
	}

	return NULL;
}

static GHashTable *procmime_get_mime_type_table(void)
{
	GHashTable *table = NULL;
	GList *cur;
	MimeType *mime_type;
	gchar **exts;

	if (!mime_type_list) {
		mime_type_list = procmime_get_mime_type_list();
		if (!mime_type_list) return NULL;
	}

	table = g_hash_table_new(g_str_hash, g_str_equal);

	for (cur = mime_type_list; cur != NULL; cur = cur->next) {
		gint i;

		mime_type = (MimeType *)cur->data;

		if (!mime_type->extension) continue;

		exts = g_strsplit(mime_type->extension, " ", 16);
		for (i = 0; exts[i] != NULL; i++)
			g_hash_table_insert(table, g_strdup(exts[i]),
					    mime_type);
		g_strfreev(exts);
	}

	return table;
}

static GList *procmime_get_mime_type_list(void)
{
	GList *list = NULL;
	FILE *fp;
	gchar buf[BUFFSIZE];
	gchar *p, *delim;
	MimeType *mime_type;

	if ((fp = fopen("/etc/mime.types", "r")) == NULL) {
		FILE_OP_ERROR("/etc/mime.types", "fopen");
		return NULL;
	}

	while (fgets(buf, sizeof(buf), fp) != NULL) {
		p = strchr(buf, '#');
		if (p) *p = '\0';
		g_strstrip(buf);

		p = buf;
		while (*p && !isspace(*p)) p++;
		if (*p) {
			*p = '\0';
			p++;
		}
		delim = strchr(buf, '/');
		if (delim == NULL) continue;
		*delim = '\0';

		mime_type = g_new(MimeType, 1);
		mime_type->type = g_strdup(buf);
		mime_type->sub_type = g_strdup(delim + 1);

		while (*p && isspace(*p)) p++;
		if (*p)
			mime_type->extension = g_strdup(p);
		else
			mime_type->extension = NULL;

		list = g_list_append(list, mime_type);
	}

	fclose(fp);

	if (!list)
		g_warning("Can't read mime.types\n");

	return list;
}
